from gensim.models.phrases import Phrases
from gensim.models.phrases import Phraser as GensimPhraser
from tqdm import tqdm
import multiprocessing as mp
import os

try:
	from .util_time import timed
	from .sorbet import sorbet
except (ModuleNotFoundError,ImportError):
	from util_time import timed
	from sorbet import sorbet

# ---[ MODEL ]------------------------------------------------------------------

class Phraser():

	@timed
	def init_phraser(self, components=False, **kwargs):
		sentences = self.all_sentences(desc='phraser_input')
		phrases = Phrases(sentences, **kwargs)
		self.phraser = GensimPhraser(phrases)
		self.phraser.components = components
		self.phraser.save(self.path+'phraser.pkl')
		del phrases
	
	def skip_phraser(self):
		self.phraser = None
		open(self.path+'phraser.pkl','wb') # empty file -> skip_phraser

	def load_phraser(self):
		path = self.path+'phraser.pkl'
		if os.path.getsize(path):
			self.phraser = GensimPhraser.load(path)
		else:
			self.phraser = None
	
	def doc_to_phrased(self, doc):
		text = self.doc_to_text(doc)
		yield from self.text_to_phrased(text)
	
	def text_to_phrased(self, text):
		if self.phraser:
			components = self.phraser.components
			sentences = self.text_to_sentences(text)
			for sentence in sentences:
				tokens = self.text_to_tokens(sentence)
				phrased = self.phraser[tokens]
				yield from phrased
				if components:
					yield from set(tokens)-set(phrased)
		else:
			yield from self.text_to_tokens(text)
	
	@timed
	def init_phrased(self, storage='disk'):
		documents = self.doc_iter()
		phrased = (list(self.doc_to_phrased(doc)) for doc in documents)
		phrased = tqdm(phrased, desc='phrased', total=len(self.meta)) # progress bar
		self.phrased = sorbet(self.path+'phrased', kind=storage).dump(phrased)

	@timed
	def init_phrased_mp(self, workers=4, chunksize=100):
		s = sorbet(self.path+'phrased').new()
		doc_id_iter = range(len(self.meta))
		doc_id_iter = tqdm(doc_id_iter,'phrased',len(self.meta))
		with mp.Pool(
					workers,
					init_phrased_worker,
					[
						self.path,
						self.get_doc_by_meta,
						self.doc_to_text,
						self.text_to_sentences,
						self.text_to_tokens,
						self.phraser==None
					]
				) as pool:
			phrased = pool.imap(phrased_worker, doc_id_iter, chunksize)
			for p in phrased:
				s.append(list(p))
		self.phrased = s.save()
	
	def skip_phrased(self):
		doc_to_text = self.doc_to_text
		text_to_tokens = self.text_to_tokens
		doc_iter = self.doc_iter
		class __phrased:
			def __iter__(self):
				return (list(text_to_tokens(doc_to_text(doc))) for doc in doc_iter())
			def delete(self): pass
		self.phrased = __phrased()
	
	def load_phrased(self, storage='disk'):
		self.phrased = sorbet(self.path+'phrased', kind=storage).load()

# ---[ MULTIPROCESSING ]--------------------------------------------------------

def init_phrased_worker(*args):
	global model
	model = Phraser()
	model.path              = args[0]
	model.get_doc_by_meta   = args[1]
	model.doc_to_text       = args[2]
	model.text_to_sentences = args[3]
	model.text_to_tokens    = args[4]
	model.skip_phraser      = args[5]
	model.meta = sorbet(model.path+'meta').load()
	if not model.skip_phraser:
		model.load_phraser()
	
def phrased_worker(doc_id):
	meta = model.meta[doc_id]
	doc = model.get_doc_by_meta(meta)
	text = model.doc_to_text(doc)
	if model.skip_phraser:
		out = model.text_to_tokens(text)
	else:
		out = []
		components = model.phraser.components
		for sen in model.text_to_sentences(text):
			tokens = model.text_to_tokens(sen)
			phrased = model.phraser[tokens]
			out.extend(phrased)
			if components:
				out.extend(set(tokens)-set(phrased)) 
	return out

# ---[ DEBUG ]-------------------------------------------------------------------

def count_digits(tokens):
	x = 0
	for token in tokens:
		for letter in token:
			if letter in "0123456789":
				x += 1
	return x

if __name__=="__main__":
	model = Phraser()
	model.path = 'model_100/'
	model.load_phraser()
	model.load_phrased()
	#
	from pprint import pprint
	#pprint(model.phrased[0])
	#
	#
	from time import time
	t0=time()
	total = 0
	aaa = {}
	aaa['total'] = 0
	if 0: # 4000 -> 35s
		for tokens in model.phrased:
			x = 0
			for token in tokens:
				for letter in token:
					if letter in "0123456789":
						x += 1
			total += x
	if 1: # 4000 -> 2p:22s
		import multiprocessing as mp
		from tqdm import tqdm
		pool = mp.Pool(processes=2)
		def agg(x):
			aaa['total'] += 1
		#results = pool.map(count_digits, tqdm(model.phrased), 100)
		#results = pool.imap_unordered(count_digits, tqdm(model.phrased), 100)
		results = pool.map_async(count_digits, tqdm(model.phrased), 100, agg)
		for x in results.get():
			total += x
	print(total)
	print(aaa['total'])
	print(f"done in {time()-t0:.02f} seconds")



from tqdm import tqdm
import pickle
from hashlib import md5
from collections import Counter

try:
	from .util_time import timed
except (ModuleNotFoundError,ImportError):
	from util_time import timed


def my_hash(text):
	"""64bit hash from text"""
	return int(md5(text.lower().encode()).hexdigest()[:16],16)

class Sentencer:

	@timed
	def init_sentencer(self, no_above=None, no_matching=None):
		self.sentencer = sen = Counter()
		self.sentencer_no_above = no_above
		self.sentencer_no_matching = no_matching
		cnt = 0
		sentences = self.all_sentences(as_tokens=False, desc='sentencer')
		for text in sentences:
			h = my_hash(text)
			sen[h] += 1
			cnt += 1
		with open(self.path+'sentencer.pkl','wb') as f:
			pickle.dump(sen, f)
			pickle.dump(no_above, f)
			pickle.dump(no_matching, f)
	
	def skip_sentencer(self):
		self.sentencer = None
		self.sentencer_no_above = None
		self.sentencer_no_matching = None
		pass # TODO
	
	def load_sentencer(self):
		with open(self.path+'sentencer.pkl','rb') as f:
			self.sentencer = pickle.load(f)
			self.sentencer_no_above = pickle.load(f)
			self.sentencer_no_matching = pickle.load(f)

	def all_sentences(self, as_tokens=True, clean=True, desc='sentences'):
		cnt = self.sentencer
		no_above = self.sentencer_no_above
		no_matching = self.sentencer_no_matching
		doc_iter = self.doc_iter()
		doc_iter = tqdm(doc_iter, desc=desc, total=len(self.meta))
		for doc in doc_iter:
			text = self.doc_to_text(doc)
			sentences = self.text_to_sentences(text)
			if cnt:
				for s in sentences:
					h = my_hash(s)
					if clean and no_above and cnt.get(h,0)>no_above: continue # skip common sentences (copyright etc)
					if clean and no_matching and no_matching.findall(s): continue # skip matching sentences
					yield self.text_to_tokens(s) if as_tokens else s
			else:
				for s in sentences:
					yield self.text_to_tokens(s) if as_tokens else s

	def explain_sentencer(self, k):
		top = self.sentencer.most_common(k)
		top_ids = set([x[0] for x in top])
		top_text = {}
		for text in self.all_sentences(as_tokens=False, clean=False):
			h = my_hash(text)
			if h in top_ids:
				top_text[h] = text
		for id,cnt in top:
			print(cnt,id,top_text.get(id,'REMOVED'))



if __name__ == "__main__":
	import data
	model = Sentencer()
	model.path = 'model_100/'
	model.load_sentencer()


from gensim.models import TfidfModel
from tqdm import tqdm
import multiprocessing as mp
import re

try:
	from .util_time import timed
	from .sorbet import sorbet
except (ModuleNotFoundError,ImportError):
	from util_time import timed
	from sorbet import sorbet

# ---[ MODEL ]------------------------------------------------------------------

class TFIDF():
	
	@timed
	def init_tfidf(self, **kwargs):
		bow = self.bow
		bow = tqdm(bow, desc='tfidf', total=len(self.meta))
		self.tfidf = TfidfModel(bow, **kwargs)
		self.tfidf.save(self.path+'tfidf.pkl')
	
	def load_tfidf(self):
		self.tfidf = TfidfModel.load(self.path+'tfidf.pkl')

	@timed
	def init_sparse(self, storage='disk'):
		sparse = (self.tfidf[bow] for bow in self.bow)
		sparse = tqdm(sparse, desc='sparse', total=len(self.meta)) # progress bar
		self.sparse = sorbet(self.path+'sparse', kind=storage).dump(sparse)
		
	def load_sparse(self, storage='disk'):
		self.sparse = sorbet(self.path+'sparse', kind=storage).load()
	
	@timed
	def init_sparse_mp(self, workers=4, chunksize=100):
		s = sorbet(self.path+'sparse').new()
		id_iter = range(len(self.meta))
		id_iter = tqdm(id_iter,'sparse',len(self.meta))
		with mp.Pool(
					workers,
					init_sparse_worker,
					[
						self.path,
					]
				) as pool:
			sparse = pool.imap(sparse_worker, id_iter, chunksize)
			for sp in sparse:
				s.append(sp)
		self.sparse = s.save()
	
	# TODO raname dfs_query ???
	def tfidf_query(self, query, exclude=None):
		dfs = []
		token_ids,tokens = self.dictionary_query(query,exclude)
		for i in token_ids:
			df = self.tfidf.dfs[i]
			dfs += [df]
		# sort
		results = list(sorted(zip(token_ids,tokens,dfs),key=lambda x:-x[2]))
		token_ids = [x[0] for x in results]
		tokens = [x[1] for x in results]
		dfs = [x[2] for x in results]
		return token_ids,tokens,dfs
			

# ---[ MULTIPROCESSING ]--------------------------------------------------------

def init_sparse_worker(*args):
	global model
	model_path = args[0]
	model = TFIDF()
	model.path = model_path
	model.bow = sorbet(model.path+'bow').load()
	model.load_tfidf()
	
def sparse_worker(doc_id):
	bow = model.bow[doc_id]
	sparse = model.tfidf[bow]
	return sparse

# ---[ DEBUG ]------------------------------------------------------------------

if __name__=="__main__":
	model = TFIDF()
	model.path = '../model_100/'
	model.load_tfidf()
	tfidf = model.tfidf
	#
	from pprint import pprint
	from itertools import islice
	#
	from model_dictionary import Dictionary
	model2 = Dictionary()
	model2.path = model.path
	model2.load_dictionary()
	d = model2.dictionary
	data = ((i,cnt,tfidf.idfs[i],d.get(i,'???')) for i,cnt in tfidf.dfs.items())
	data = sorted(data,key=lambda x:x[1],reverse=True)
	for i,cnt,idf,token in islice(data,40):
		print(i,cnt,
			round(idf,1),
			token,
			sep='\t')



from gensim.models import LsiModel
from tqdm import tqdm
from array import array
import multiprocessing as mp

try:
	from .util_time import timed
	from .sorbet import sorbet
except (ModuleNotFoundError,ImportError):
	from util_time import timed
	from sorbet import sorbet

# ---[ MODEL ]------------------------------------------------------------------

class LSI():
	
	@timed
	def init_lsi(self, **kwargs):
		corpus = self.sparse
		corpus = tqdm(corpus, desc='lsi_input', total=len(self.sparse)) # progress bar
		self.lsi = LsiModel(corpus, **kwargs)
		self.lsi.save(self.path+'lsi.pkl')
	
	def load_lsi(self):
		self.lsi = LsiModel.load(self.path+'lsi.pkl')

	# TODO use sparse_to_dense
	@timed
	def init_dense(self, storage='disk'):
		corpus = self.sparse
		#num_topics = self.lsi.num_topics
		#dense = (self.lsi[c] for c in corpus)
		#dense = ([dict(d).get(i,0) for i in range(num_topics)] for d in dense)
		dense = (self.sparse_to_dense(c) for c in corpus)
		dense = (array('f',d) for d in dense) # uses 4x less memory on disk
		dense = tqdm(dense, desc='dense', total=len(corpus)) # progress bar
		self.dense = sorbet(self.path+'dense', kind=storage).dump(dense)
		
	def load_dense(self, storage='disk'):
		self.dense = sorbet(self.path+'dense', kind=storage).load()
	
	# TODO gensim.matutils.sparse2full
	def sparse_to_dense(self, sparse):
		dense = self.lsi[sparse]
		dense = [dict(dense).get(i,0) for i in range(self.lsi.num_topics)]
		return dense

	@timed
	def init_dense_mp(self, workers=4, chunksize=100):
		s = sorbet(self.path+'dense').new()
		id_iter = range(len(self.meta))
		id_iter = tqdm(id_iter,'dense',len(self.meta))
		with mp.Pool(
					workers,
					init_dense_worker,
					[
						self.path,
					]
				) as pool:
			dense = pool.imap(dense_worker, id_iter, chunksize)
			for d in dense:
				s.append(d)
		self.dense = s.save()

# ---[ MULTIPROCESSING ]--------------------------------------------------------

def init_dense_worker(*args):
	global model
	model_path = args[0]
	model = LSI()
	model.path = model_path
	model.sparse = sorbet(model.path+'sparse').load()
	model.load_lsi()

# TODO gensim.matutils.sparse2full
def dense_worker(doc_id):
	sparse = model.sparse[doc_id]
	dense = model.lsi[sparse]
	dense = [dict(dense).get(i,0) for i in range(model.lsi.num_topics)]
	dense = array('f',dense)
	return dense


# ---[ DEBUG ]------------------------------------------------------------------

if __name__=="__main__":
	model = LSI()
	model.path = 'model_100/'
	model.load_lsi()
	model.load_dense()
	#
	print(model.dense[0])

